#include <Arduino.h>
#include <FlexCAN_T4.h>

//Function Declaration
void sendEnable(); // function to turn the inverter on DON'T FORGET TO EDIT THE ID OF THE DRIVE
void sendERPM(); // function to change the speed DON'T FORGET TO EDIT THE ID OF THE DRIVE

/* CAN variables */
FlexCAN_T4<CAN1, RX_SIZE_256, TX_SIZE_16> can_lv;  /* LV CAN bus is connected to controller 1 */
FlexCAN_T4<CAN2, RX_SIZE_256, TX_SIZE_16> can_mc;  /* Motorcontroller/Invertor CAN bus is connected to controller 2 */
#define CAN_LV_BAUD 1000000                        /* Data rate of the LV CAN = 1Mbps */
#define CAN_MC_BAUD 500000                         /* Data rate of the MC CAN = 500kbps */
uint64_t time_last_can_mc_send = 0;                /* Timestamp of last message that was sent to motorcontrollers (because they have timeout safety shit) */
uint64_t time_last_apps = 0;                       /* Timestamp of last message for APPS values */
uint64_t time_last_rtd_state = 0;                  /* Timestamp of last message for Boot Sequence State stuff */

/* CAN heartbeat stuff */
uint64_t time_last_heartbeat = 0;  /* Timestamp of last heartbeat message that was sent */
#define HEARTBEAT_PERIOD 100       /* Send heartbeat message every 100ms */
#define CAN_ID_HEARTBEAT 0x080     /* CAN ID of heartbeat messages (see spreadsheet) */
#define CAN_MODULE_ID 0x05         /* Module ID of this system (see spreadsheet) */

/* CAN config continued */
#define CAN_ID_APPS      0x110  /* CAN ID of APPS values (not using Module ID of APPS) */
#define CAN_ID_BOOT_SEQ  0x090  /* CAN ID of Boot Sequence Status (for RTD) values (using Module ID of ECU for safety) */

/* Torque Controller variables */
uint8_t apps_value = 0;  /* Value of the APPS sensor (0-255) */
char rtd_state = 0;      /* Whether or not RTD is actived */


void handle_can_lv_message(const CAN_message_t &msg) {
	/* Callback for LV CAN messages */

	if ((msg.id & 0xff0) == CAN_ID_APPS) {
		/* APPS value */
		apps_value = msg.buf[0];
		time_last_apps = millis();
	}
	else if ((msg.id & 0xfff) == CAN_ID_BOOT_SEQ) {
		/* Boot Sequence Status */
		char bs = msg.buf[0];
		rtd_state = (bs == 5) ? 0b1 : 0b0;
		time_last_rtd_state = millis();
	}
}

void handle_can_mc_message(const CAN_message_t &msg) {
	/* Callback for MC CAN messages */
}


void can_setup() {
	/* Init LV CAN */
	can_lv.begin();
	can_lv.setBaudRate(CAN_LV_BAUD);
	can_lv.setMaxMB(16);
	can_lv.enableFIFO();
	can_lv.enableFIFOInterrupt();
	can_lv.onReceive(handle_can_lv_message);

	/* Init MC CAN */
	can_mc.begin();
	can_mc.setBaudRate(CAN_MC_BAUD);
	can_mc.setMaxMB(16);
	can_mc.enableFIFO();
	can_mc.enableFIFOInterrupt();
	can_mc.onReceive(handle_can_mc_message);
}


// ID bij ons was 79 = 0x4F

void sendEnable() {
    CAN_message_t msg_DriveEnable;
    msg_DriveEnable.flags.extended = 1;
    msg_DriveEnable.id = 0x24 << 8 | 0x4F; // INVERTER ID INSTELLEN VOLGENS VOLGEND FORMAT: 0x24 << 8 | 0x"ID"
    msg_DriveEnable.buf[0] = 1;
    msg_DriveEnable.buf[1] = 255;
    msg_DriveEnable.buf[2] = 0xFF;
    msg_DriveEnable.buf[3] = 0xFF;
    msg_DriveEnable.buf[4] = 0xFF;
    msg_DriveEnable.buf[5] = 0xFF;
    msg_DriveEnable.buf[6] = 0xFF;
    msg_DriveEnable.buf[7] = 0xFF;
    Can0.write(msg_DriveEnable);
    Serial.println(); 
}


void sendERPM() {
  if ( millis() - time_last_update_can > 100 ) {
    
    sendEnable();
    Serial.print("Enable");

    CAN_message_t msg_ERPM;
    msg_ERPM.flags.extended = 1;
    msg_ERPM.id = 0x1C << 8 | 0x4F; // INVERTER ID INSTELLEN VOLGENS VOLGEND FORMAT: 0x1C << 8 | 0x"ID"
    msg_ERPM.buf[0] = (speed >> 24);
    msg_ERPM.buf[1] = (speed >> 16) & 0xFF;
    msg_ERPM.buf[2] = (speed >> 8) & 0xFF;
    msg_ERPM.buf[3] = speed & 0xFF;
    Can0.write(msg_ERPM);
    Serial.println("ERPM");
    printMessage(msg_ERPM);
    time_last_update_can = millis();
  }
}

/* Subroutine to check if new heartbeat message needs to be sent and transmits it if needed */
void check_heartbeat() {
	/* Return if not needed yet */
	if (millis() < time_last_heartbeat + HEARTBEAT_PERIOD) return;

	/* Send heartbeat */
	CAN_message_t m;
	m.id = CAN_ID_HEARTBEAT | CAN_MODULE_ID;
	m.buf[0] = HEARTBEAT_STATUS_OK;
	can_lv.write(m);

	/* Update stats */
	time_last_heartbeat = millis();
}

/* Subroutine to update invertors */
void update_drives() {
	/* TODO: Return if not needed yet */
	//if (millis() < time_last_heartbeat + HEARTBEAT_PERIOD) return;

	/* Update speed */
	uint8_t speed = apps_value;

	/* Check if data is recent enough */
	//if (millis() > time_last_apps 

	/* Send heartbeat */
	CAN_message_t m;
	m.id = CAN_ID_HEARTBEAT | CAN_MODULE_ID;
	m.buf[0] = HEARTBEAT_STATUS_OK;
	can_lv.write(m);

	/* Update stats */
	time_last_heartbeat = millis();
}

void setup() {
	/* Init CAN */
	can_setup();
}

void loop() {
	/* Main event loop */
	can_lv.events();    /* Handle CAN LV events */
	can_mc.events();    /* Handle CAN MC events */
	check_heartbeat();  /* Do heartbeat stuffz */
	update_drives();    /* Do motorcontroller shit */
}

